import json

from django.contrib import admin
from django.utils.html import format_html

# Register your models here.
from .models.core import Customer, Card, Charge, Source, Refund
from .models.schedule import Schedule, Occurrence, ChargeSchedule
from .models.event import Event


class CardInline(admin.TabularInline):

    model = Card

    fields = (
        "id",
        "livemode",
        "deleted",
        "last_digits",
    )

    # classes = [
    #     "collapse",
    # ]

    extra = 0

    show_change_link = True

    can_delete = False

    def has_change_permission(self, request, obj=None):
        return False


class ChargeScheduleInline(admin.TabularInline):

    model = ChargeSchedule

    fields = [
        "id",
        "livemode",
        "human_amount",
        "currency",
        "card",
        "default_card",
        "schedule",
        "schedule_status",
        "next_occurrence_on",
        "schedule_in_words",
    ]

    readonly_fields = [
        "human_amount",
        "schedule",
        "schedule_status",
        "next_occurrence_on",
        "schedule_in_words",
    ]

    extra = 0
    can_delete = False
    show_change_link = True

    def has_change_permission(self, request, obj=None):
        return False

    def schedule(self, obj=None):
        if obj:
            return obj.schedule

    def schedule_status(self, obj=None):
        if obj:
            return obj.schedule.status

    def next_occurrence_on(self, obj=None):
        if obj and obj.schedule.next_occurrences_on:
            return obj.schedule.next_occurrences_on[0]

    def schedule_in_words(self, obj=None):
        if obj:
            return obj.schedule.in_words


class ChargeInline(admin.TabularInline):

    model = Charge
    extra = 0
    can_delete = False

    show_change_link = True

    fields = [
        "id",
        "livemode",
        "status",
        "human_amount",
        "currency",
        "refunded_amount",
        "schedule",
        "card",
        "source",
    ]

    # classes = [
    #     "collapse",
    # ]

    readonly_fields = ("human_amount",)

    def has_change_permission(self, request, obj=None):
        return False


@admin.register(Customer)
class CustomerAdmin(admin.ModelAdmin):

    search_fields = (
        "user__email",
        "user__username",
        "id",
    )

    list_display = (
        "id",
        "deleted",
        "user",
        "livemode",
        "date_created",
        "date_updated",
    )

    readonly_fields = ("uid",)

    inlines = [
        CardInline,
        ChargeScheduleInline,
        ChargeInline,
    ]

    list_filter = (
        "livemode",
        "date_created",
        "date_updated",
        "deleted",
    )


@admin.register(Card)
class CardAdmin(admin.ModelAdmin):
    list_display = (
        "customer",
        "id",
        "livemode",
        "last_digits",
        "deleted",
        "date_created",
        "date_updated",
    )
    list_filter = (
        "livemode",
        "deleted",
        "date_created",
        "date_updated",
    )

    readonly_fields = (
        "uid",
        "id",
        "livemode",
        "customer",
        "financing",
        "fingerprint",
    )
    search_fields = ("customer__user__email", "customer__user__username", "id")

    inlines = [
        ChargeScheduleInline,
        ChargeInline,
    ]


@admin.register(Refund)
class RefundAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "livemode",
        "date_created",
        "charge",
        "amount",
        "currency",
        "funding_amount",
        "funding_currency",
        "voided",
    )
    list_filter = (
        "livemode",
        "date_created",
        "date_updated",
        "voided",
    )
    search_fields = [
        "charge__id",
    ]

    def has_change_permission(self, request, obj=None):
        return False


@admin.register(Event)
class EventAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "livemode",
        "event_type",
        "date_created",
    )
    readonly_fields = (
        "uid",
        "event_object",
        "content_type",
        "object_id",
    )
    list_filter = ("livemode", "date_created", "date_updated", "event_type")

    fields = (
        "id",
        "livemode",
        "event_object",
        "event_type",
        "event_data",
        "uid",
        "content_type",
        "object_id",
    )

    def event_data(self, obj=None):
        return format_html("<pre>{}</pre>", json.dumps(obj.data, indent=4))

    def has_change_permission(self, request, obj=None):
        return False


class RefundInline(admin.TabularInline):
    model = Refund

    extra = 0

    fields = ["id", "livemode", "human_amount", "currency", "voided"]
    readonly_fields = ("human_amount",)

    def has_change_permission(self, request, obj=None):
        return False

    show_change_link = True


@admin.register(Charge)
class ChargeAdmin(admin.ModelAdmin):
    search_fields = ("id",)

    def has_change_permission(self, request, obj=None):
        return False

    list_display = (
        "id",
        "livemode",
        "date_created",
        "date_updated",
        "status",
        "currency",
        "amount",
        "refunded_amount",
        "refundable",
        # "authorize_uri",
        # "authorized",
        # "capturable",
        # "capture",
        # "card",
        "source_type",
        "schedule",
        # "customer",
        "description",
        # "disputable",
        # "expired",
        # "expired_at",
        # "expires_at",
        # "failure_code",
        "failure_message",
        # "fee",
        # "fee_vat",
        # "funding_amount",
        # "funding_currency",
        # "interest",
        # "interest_vat",
        # "ip",
        # "net",
        # "paid",
        # "paid_at",
        # "reversed",
        "source",
        "voided",
        "zero_interest_installments",
    )
    list_filter = (
        "livemode",
        "date_created",
        "date_updated",
        "authorized",
        "capturable",
        "capture",
        "disputable",
        "expired",
        "expired_at",
        "expires_at",
        "paid",
        "paid_at",
        "refundable",
        "reversed",
        "voided",
        "zero_interest_installments",
    )

    inlines = [
        RefundInline,
    ]

    readonly_fields = ("uid", "source_type")

    def source_type(self, obj=None):

        if obj and obj.card:
            return f"card ends with {obj.card.last_digits}"

        if obj and obj.source:
            return obj.source.get_type_display()


@admin.register(Source)
class SourceAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "amount",
        "bank",
        "barcode",
        "charge_status",
        "currency",
        "email",
        "flow",
        "installment_term",
        "mobile_number",
        "name",
        "phone_number",
        "platform_type",
        "receipt_amount",
        "store_id",
        "store_name",
        "terminal_id",
        "type",
        "zero_interest_installments",
    )
    list_filter = ("zero_interest_installments",)
    search_fields = (
        "id",
        "name",
    )
    readonly_fields = ("uid",)

    def has_change_permission(self, request, obj=None):
        return False


class OccurrenceInline(admin.TabularInline):
    model = Occurrence

    extra = 0

    show_change_link = True

    def has_change_permission(self, request, obj=None):
        return False


@admin.register(Schedule)
class ScheduleAdmin(admin.ModelAdmin):
    list_display = (
        "status",
        "id",
        "livemode",
        "date_created",
        "date_updated",
        "deleted",
        "active",
        "every",
        "in_words",
        "period",
        "start_on",
    )
    list_filter = (
        "livemode",
        "date_created",
        "date_updated",
        "deleted",
        "active",
    )

    fields = [
        "id",
        "livemode",
        "data",
        "deleted",
        "active",
        "charge",
        "charge_amount",
        "end_on",
        "ended_at",
        "every",
        "in_words",
        "next_occurrences_on",
        "period",
        "start_on",
        "status",
    ]

    inlines = [
        OccurrenceInline,
    ]

    def has_change_permission(self, request, obj=None):
        return False

    def charge_amount(self, obj=None):
        if obj and obj.charge:
            return f"{obj.charge.human_amount} {obj.charge.currency}"


@admin.register(Occurrence)
class OccurrenceAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "livemode",
        "date_created",
        "date_updated",
        "message",
        "processed_at",
        "retry_on",
        "schedule",
        "schduled_on",
        "status",
    )
    list_filter = (
        "livemode",
        "date_created",
        "date_updated",
        "processed_at",
        "retry_on",
        "schedule",
        "schduled_on",
    )

    def has_change_permission(self, request, obj=None):
        return False


@admin.register(ChargeSchedule)
class ChargeScheduleAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "livemode",
        "date_created",
        "date_updated",
        "amount",
        "currency",
        "card",
        "customer",
        "default_card",
    )
    list_filter = (
        "livemode",
        "date_created",
        "date_updated",
        "default_card",
    )

    def has_change_permission(self, request, obj=None):
        return False



from .base import OmiseBaseModel, OmiseMetadata
from .choices import Currency, ChargeStatus, ChargeSourceType, SourceFlow
from .managers import NotDeletedManager

import uuid

from django.apps import apps
from django_omise.omise import omise

from django.conf import settings
from django.contrib.auth import get_user_model

from django.db import models

from django.urls import reverse_lazy, reverse
from django.utils.translation import gettext_lazy as _

from typing import TYPE_CHECKING, Optional, Dict

if TYPE_CHECKING:
    User = get_user_model()


class Customer(OmiseBaseModel, OmiseMetadata):

    """
    A card representing Omise Customer object.

    Official documentation: https://www.omise.co/customers-api
    """

    NON_DEFAULT_FIELDS = OmiseBaseModel.NON_DEFAULT_FIELDS + [
        "charges",
        "user",
        "charge_schedules",
    ]

    omise_class = omise.Customer

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name="omise_customers",
    )

    deleted = models.BooleanField(
        default=False, help_text=_("Whether this customer was deleted.")
    )

    default_card = models.ForeignKey(
        "Card", on_delete=models.SET_NULL, blank=True, null=True, related_name="+"
    )

    description = models.TextField(blank=True)

    email = models.EmailField(blank=True)

    objects = NotDeletedManager()

    def __str__(self) -> str:
        if self.user:
            return f"Omise{self.__class__.__name__}: {str(self.user)}"
        return f"Omise{self.__class__.__name__}: {self.id}"

    def add_card(self, token: omise.Token) -> "Card":
        """
        Add a new card to the user

        :param token: The token retrieved from Omise API.

        :returns: A new card instance.
        """
        omise_customer = omise.Customer.retrieve(self.id)
        omise_customer.update(card=token.id)

        card = token.card

        new_card = Card.update_or_create_from_omise_object(
            omise_object=card,
        )

        new_card.customer = self
        new_card.save()

        return new_card

    @property
    def schedules(self):
        """
        Get a list of all charge schedules belonging to this customer.

        :returns: A list of django_omise.models.schedule.Schedule
        """
        schedule_model = apps.get_model(app_label="django_omise", model_name="Schedule")
        schedule_ids = self.charge_schedules.all().values_list("schedule", flat=True)
        schedules = [schedule_model.objects.get(pk=pk) for pk in schedule_ids]
        return schedules

    def remove_card(self, card: "Card") -> None:
        """
        Remove the card from the customer.

        :param card: The card to be removed.

        :returns: None
        """
        self.cards.filter(id=card.id).delete()

    def charge_with_card(
        self,
        amount: int,
        currency: Currency.choices,
        card: "Card",
        return_uri: str = None,
        metadata: dict = None,
    ) -> "Charge":
        """
        Charge the customer with provided card.

        :returns: An instace of Charge object.
        """
        return Charge.charge(
            amount=amount,
            currency=currency,
            card=card,
            return_uri=return_uri,
            metadata=metadata,
        )

    @classmethod
    def get_or_create(cls, user: "User") -> "Customer":
        """
        Get or create django_omise customer.o

        :param user: The user model instance to create a customer.
        :returns: A tuple of Customer instance and whether the object is newly created.
        """
        livemode = settings.OMISE_LIVE_MODE

        try:
            return (
                Customer.objects.get(user=user, livemode=livemode, deleted=False),
                False,
            )
        except Customer.DoesNotExist:

            customer_attributes = {}

            if user.email:
                customer_attributes["email"] = user.email

            omise_customer = cls.omise_class.create(**customer_attributes)

            if omise_customer.livemode != livemode:
                raise ValueError(
                    f"The API livemode inconsistent. API livemode: {omise_customer.livemode}. settings.OMISE_LIVE_MODE: {livemode}"
                )

            return (
                Customer.objects.create(
                    id=omise_customer.id,
                    user=user,
                    livemode=omise_customer.livemode,
                ),
                True,
            )


class Card(OmiseBaseModel):
    """
    A class representing Omise's Card object.

    Official documentation: https://www.omise.co/cards-api
    """

    omise_class = omise.Card

    NON_DEFAULT_FIELDS = OmiseBaseModel.NON_DEFAULT_FIELDS + [
        "charges",
        "customer",
        "charge_schedules",
    ]

    customer = models.ForeignKey(
        Customer,
        blank=True,
        null=True,
        related_name="cards",
        on_delete=models.PROTECT,
    )

    name = models.CharField(max_length=255, blank=True)
    phone_number = models.CharField(max_length=255, blank=True)

    postal_code = models.CharField(max_length=255, blank=True)
    state = models.CharField(max_length=255, blank=True)
    street1 = models.CharField(max_length=255, blank=True)
    street2 = models.CharField(max_length=255, blank=True)
    tokenization_method = models.CharField(max_length=255, blank=True)
    first_digits = models.CharField(max_length=6, blank=True)
    last_digits = models.CharField(max_length=4)
    bank = models.CharField(max_length=255, blank=True)
    brand = models.CharField(max_length=255, blank=True)
    city = models.CharField(max_length=255, blank=True)
    country = models.CharField(max_length=255, blank=True)

    expiration_month = models.CharField(max_length=2, blank=True)
    expiration_year = models.CharField(max_length=4, blank=True)

    fingerprint = models.TextField(blank=True)

    financing = models.CharField(max_length=10, blank=True)

    deleted = models.BooleanField(
        default=False, help_text=_("Whethe this card was deleted.")
    )

    objects = NotDeletedManager()

    def __str__(self) -> str:
        return f"Omise{self.__class__.__name__}: {self.last_digits}"

    def delete(self):
        omise_customer = self.customer.get_omise_object()
        card = omise_customer.cards.retrieve(self.id)
        card.destroy()

        self.deleted = True
        self.save()


class Charge(OmiseBaseModel, OmiseMetadata):
    """
    A class representing Omise Charge object.

    Official documentation: https://www.omise.co/charges-api
    """

    omise_class = omise.Charge

    status = models.CharField(
        max_length=10,
        choices=ChargeStatus.choices,
        help_text=_("Status of this charge."),
    )

    amount = models.IntegerField(
        help_text="Charge amount in the smallest unit of charge currency"
    )
    authorize_uri = models.CharField(
        max_length=255,
        help_text=_("URI for payment authorization (e.g. 3-D Secure)."),
        blank=True,
    )

    authorized = models.BooleanField(
        blank=True, null=True, help_text=_("Whether charge has been authorized.")
    )
    capturable = models.BooleanField(
        blank=True,
        null=True,
        help_text=_("Whether charge is able to be captured (paid)."),
    )
    capture = models.BooleanField(
        blank=True,
        null=True,
        help_text=_("Whether charge is set to be automatically captured (paid)."),
    )

    card = models.ForeignKey(
        Card,
        blank=True,
        null=True,
        on_delete=models.PROTECT,
        related_name="charges",
        help_text=_("Card that was charged (if card was charged)."),
    )

    currency = models.CharField(
        max_length=3,
        choices=Currency.choices,
    )

    customer = models.ForeignKey(
        Customer,
        blank=True,
        null=True,
        on_delete=models.PROTECT,
        related_name="charges",
        help_text=_("Customer associated with charge."),
    )

    description = models.TextField(blank=True)
    disputable = models.BooleanField(blank=True, null=True)

    expired = models.BooleanField(
        blank=True, null=True, help_text=_("Whether this charge has expired.")
    )

    expired_at = models.DateTimeField(blank=True, null=True)
    expires_at = models.DateTimeField(blank=True, null=True)

    failure_code = models.CharField(max_length=255, blank=True)
    failure_message = models.TextField(blank=True)

    fee = models.IntegerField()
    fee_vat = models.IntegerField()

    funding_amount = models.IntegerField(
        blank=True,
        null=True,
        help_text=_(
            "For multi-currency charges, amount after exchange into account funding currency."
        ),
    )

    funding_currency = models.CharField(
        max_length=3,
        choices=Currency.choices,
    )

    interest = models.IntegerField(
        blank=True,
        null=True,
        help_text=_(
            "Interest paid by customer or merchant over the course of an installment term."
        ),
    )

    interest_vat = models.IntegerField(
        blank=True,
        null=True,
        help_text=_("Value-added Tax (VAT) applied to interest."),
    )

    ip = models.CharField(
        max_length=255,
        blank=True,
        help_text=_(
            "IP address provided to Omise at charge creation. May be IPv4 or IPv6."
        ),
    )

    net = models.IntegerField(
        help_text=_("funding_amount after fees, interest, and VAT deducted.")
    )

    paid = models.BooleanField(help_text=_("Whether charge has been captured (paid)."))
    paid_at = models.DateTimeField(blank=True, null=True)
    refundable = models.BooleanField(
        help_text=_("Whether charge is able to be refunded.")
    )

    refunded_amount = models.IntegerField(
        blank=True,
        null=True,
        help_text=_("Refunded amount in smallest unit of currency."),
    )

    return_uri = models.TextField(
        blank=True,
        help_text=_(
            "URI to which customer is redirected after 3-D Secure card authorization or other redirect-based authorization."
        ),
    )

    reversed = models.BooleanField(
        blank=True,
        null=True,
        help_text=_("Whether charge authorization is able to be reversed."),
    )

    schedule = models.ForeignKey(
        "django_omise.Schedule",
        on_delete=models.PROTECT,
        related_name="charges",
        blank=True,
        null=True,
    )

    source = models.ForeignKey(
        "Source",
        blank=True,
        null=True,
        on_delete=models.PROTECT,
        related_name="charges",
    )

    voided = models.BooleanField(
        blank=True,
        null=True,
        help_text=_(
            "Whether, in the case of a refund, charge was voided instead. Charges are voided if refund is processed before settlement time."
        ),
    )

    zero_interest_installments = models.BooleanField(
        blank=True,
        null=True,
        help_text=_(
            "Whether merchant absorbs the interest for installment payments; must match value in associated source."
        ),
    )

    @property
    def human_amount(self) -> float:
        if not self.pk:
            return 0
        if self.currency == Currency.JPY:
            return f"{self.amount:,.2f}"
        return f"{self.amount / 100:,.2f}"

    @classmethod
    def charge(
        cls,
        amount: int,
        currency: Currency.choices,
        token: Optional[omise.Token] = None,
        card: Optional[Card] = None,
        source: Optional[Dict] = None,
        return_uri: str = None,
        metadata: dict = None,
    ) -> "Charge":
        """
        Charge the customer with provided payment_method.

        :param amount: The amount to charge in the smallest unit.
        :param currency: The currency to charge.
        :param token: The token to charge.
        :param card: The card to charge.
        :param source: The source to charge.
        :param return_uri optional: The return uri.
        :param metadata optional: The charge metadata.
        :returns: An instace of Charge object.
        """
        if [token, card, source].count(None) == 3:
            raise ValueError("At least a token, a card, or a source is required")

        if [token, card, source].count(None) != 2:
            raise ValueError("Only one of token, card, or source must be specified")

        uid = uuid.uuid4()

        host = settings.OMISE_CHARGE_RETURN_HOST
        if return_uri is None:
            return_uri = f'https://{host}{reverse("django_omise:return_uri", kwargs={"uid": uid})}'

        if metadata is None:
            metadata = {}

        charge_details = {}

        if token is not None:
            if type(token) == str:
                charge_details["card"] = token
            else:
                charge_details["card"] = token.id

        if card is not None:
            charge_details["card"] = card.id
            charge_details["customer"] = card.customer.id

        if source is not None:
            charge_details["source"] = source

        charge = omise.Charge.create(
            amount=int(amount),
            currency=currency,
            metadata=metadata,
            return_uri=return_uri,
            **charge_details,
        )

        return cls.update_or_create_from_omise_object(omise_object=charge, uid=uid)


class Source(OmiseBaseModel):
    """
    A class representing Omise Source object.

    Official documentation: https://www.omise.co/sources-api
    """

    omise_class = omise.Source

    NON_DEFAULT_FIELDS = OmiseBaseModel.NON_DEFAULT_FIELDS + [
        "charges",
    ]

    amount = models.IntegerField(
        help_text=_("Source amount in smallest unit of source currency")
    )

    bank = models.CharField(max_length=255, blank=True)

    barcode = models.CharField(
        max_length=255,
        blank=True,
        help_text=_("Source amount in smallest unit of source currency"),
    )

    charge_status = models.CharField(
        max_length=10,
        choices=ChargeStatus.choices,
        help_text=_("Status of charge created using this source (if any)"),
    )

    currency = models.CharField(
        max_length=3,
        choices=Currency.choices,
    )

    email = models.CharField(
        max_length=255, blank=True, help_text=_("The payer's email")
    )

    flow = models.CharField(
        max_length=50,
        help_text=_(
            "The payment flow payers need to go through to complete the payment. One of redirect, offline, or app_redirect"
        ),
        choices=SourceFlow.choices,
    )

    installment_term = models.IntegerField(
        blank=True, null=True, help_text=_("Installment term in months")
    )
    mobile_number = models.CharField(max_length=50, blank=True)
    name = models.CharField(max_length=255, blank=True, help_text=_("Payer's name"))
    phone_number = models.CharField(max_length=50, blank=True)
    platform_type = models.CharField(
        max_length=50,
        blank=True,
        help_text=_(
            "Platform from which payer is making a payment. One of WEB, IOS, or ANDROID"
        ),
    )

    receipt_amount = models.IntegerField(blank=True, null=True)
    store_id = models.CharField(max_length=255, blank=True)
    store_name = models.CharField(max_length=255, blank=True)
    terminal_id = models.CharField(max_length=255, blank=True)

    type = models.CharField(max_length=255, choices=ChargeSourceType.choices)

    zero_interest_installments = models.BooleanField(blank=True, null=True)

    @property
    def source_type_name(self):
        source_types = {
            "truemoney": _("TrueMoney Wallet"),
            "internet_banking_bay": _("Internet Banking Krungsri Bank"),
            "internet_banking_bbl": _("Internet Banking Bangkok Bank"),
            "internet_banking_ktb": _("Internet Banking Krungthai Bank"),
            "internet_banking_scb": _("Internet Banking SCB Bank"),
            "promptpay": _("Promptpay"),
        }
        return source_types[self.type]

    @property
    def human_amount(self) -> float:
        if not self.pk:
            return 0
        if self.currency == Currency.JPY:
            return f"{self.amount:,.2f}"
        return f"{self.amount / 100:,.2f}"


class Refund(OmiseBaseModel, OmiseMetadata):

    charge = models.ForeignKey(
        Charge,
        related_name="refunds",
        on_delete=models.PROTECT,
    )

    amount = models.IntegerField(
        help_text="Refund amount in smallest unit of charge currency.",
    )

    currency = models.CharField(
        max_length=3,
        choices=Currency.choices,
    )

    funding_amount = models.IntegerField(
        blank=True,
        null=True,
        help_text=_(
            "For multi-currency charges, amount after exchange into account funding currency."
        ),
    )

    funding_currency = models.CharField(
        max_length=3,
        choices=Currency.choices,
    )

    voided = models.BooleanField()

    @property
    def human_amount(self) -> float:
        if not self.pk:
            return 0
        if self.currency == Currency.JPY:
            return f"{self.amount:,.2f}"
        return f"{self.amount / 100:,.2f}"



from django_omise.models.event import EventType, Event
from django_omise.models.core import Charge
from django_omise.omise import omise

from django_omise.utils.core_utils import update_or_create_from_omise_object


from typing import Dict


def pre_event_handle(omise_event: omise.Event, raw_event: Dict):
    """
    Perform additional actions on Omise Event object before handling with the webhook view.

    :params omise_event: The Omise Event object
    :param raw_event: The event data as received with webhook view

    :returns: None
    """
    if omise_event.key == EventType.REFUND_CREATE.value:
        charge = Charge.objects.get(pk=omise_event.data.charge)
        charge.reload_from_omise()

    if omise_event.key.startswith("schedule."):
        update_or_create_from_omise_object(
            omise_object=omise_event.data,
            ignore_fields=[
                "occurrences",
            ],
        )


def post_event_handle(omise_event: omise.Event, event_object: Event, raw_event: Dict):
    """
    Perform additional actions on Omise Event object after handling with the webhook view.

    :params omise_event: The Omise Event object
    :param event_object: The new event object in the database
    :param raw_event: The event data as received with webhook view

    :returns: None
    """

    if omise_event.key.startswith("charge."):

        schedule_id = raw_event.get("data", {}).get("schedule", None)
        omise_schedule = None
        schedule = None

        if schedule_id:
            omise_schedule = omise.Schedule.retrieve(schedule_id)
            schedule = update_or_create_from_omise_object(omise_object=omise_schedule)

        charge = event_object.event_object
        charge.schedule = schedule
        charge.save()

    if omise_event.key.startswith("customer."):

        customer = event_object.event_object
        for schedule in customer.schedules:
            schedule.reload_from_omise()



from .forms import CheckoutWithCardsForm, CheckoutForm
from .models.core import Customer, Charge, Card
from .models.choices import Currency, ChargeStatus, SourceFlow, ChargeSourceType
from .omise import omise

from django.contrib import messages
from django.shortcuts import redirect
from django.urls import reverse
from django.utils.translation import gettext as _
from django.views.generic import FormView

from typing import Dict, Optional


class CheckoutMixin(FormView):
    form_class = CheckoutForm

    def charge(self, charge_type_details: Dict) -> Charge:
        """
        Created a charge with selected payment method.

        :param charge_type_details: Dict of charge type details.
                                    E.g. {'card': core.Card}
                                    or {'token': omise.Token}
                                    or {'source': {'type': 'internet_banking_bbl'}}

        :returns: A new Charge instance.
        """
        charge_details = self.get_charge_details()
        charge_kwargs = self.get_charge_kwargs()

        amount = charge_details["amount"]
        currency = charge_details["currency"]

        charge = Charge.charge(
            amount=amount,
            currency=currency,
            **charge_type_details,
            **charge_kwargs,
        )

        return charge

    def get_form_kwargs(self, *args, **kwargs) -> Dict:
        """
        Overwrite this method to provide data for the form.

        :returns: A dictionary in the form of {'user': auth.User, 'amount': int, 'currency': Currency}
        """
        kwargs = super().get_form_kwargs(*args, **kwargs)
        kwargs["user"] = self.request.user
        return {**kwargs, **self.get_charge_details()}

    def get_charge_details(self) -> Dict[int, Currency]:
        """
        Overwrite this method to provide the charge amount and currency.

        Read more on the amount and currency here https://www.omise.co/currency-and-amount

        :returns: Dictionary of charge amount in the smallest unit (int) and the 3-digit currency code. {"amount": int, "currency": str}
        """
        raise NotImplementedError("Must rewrite get_charge_details method.")

    def get_omise_customer(self) -> Optional[Customer]:
        """
        Get the user to charge. Neccessary for charging with existing card.
        Overwrite this method to provide different way to obtain the customer instance.

        :returns: An instance of Customer if the user is logged in. None otherwise.
        """
        user = self.request.user

        if user.is_authenticated == False:
            return None

        customer, created = Customer.get_or_create(user=self.request.user)
        return customer

    def get_charge_kwargs(self) -> Dict:
        """
        Overwrite this method to add charge parameters.

        E.g. You could return {'metadata': {'successful_url': ..., 'failed_url': ...}} to redirect to specific view
        when a user complete actions from pending charges.

        :return: A dictionary of charge parameters
        """
        return dict()

    def form_valid(self, form):

        charge_type_details = form.get_charge_type_details()

        try:
            charge = self.charge(charge_type_details=charge_type_details)
        except omise.errors.InvalidChargeError as e:
            messages.error(
                self.request,
                f"Payment failed: {str(e)}",
            )
            return super().form_invalid(form)

        self.process_charge_and_form(charge, form)

        if charge.status == ChargeStatus.FAILED:
            messages.error(
                self.request,
                f"Payment failed: {charge.failure_message}",
            )
            return super().form_invalid(form)

        if charge.status == ChargeStatus.PENDING:
            if not charge.source:
                return redirect(charge.authorize_uri)

            if charge.source.flow != SourceFlow.OFFLINE:
                return redirect(charge.authorize_uri)

            if charge.source.type == ChargeSourceType.PROMPTPAY:
                return redirect(
                    reverse(
                        "django_omise:promptpay_checkout",
                        kwargs={"pk": charge.id},
                    )
                )

        if charge.status == ChargeStatus.SUCCESSFUL:
            messages.success(self.request, _("Payment successful."))

        return super().form_valid(form)

    def process_charge_and_form(self, charge: Charge, form):
        """
        Overwrite this method to process the new charge object with the form.

        :params charge: The new Charge object.
        :params form: The form instance
        """
        pass


class CheckoutWithCardsMixin(FormView):
    form_class = CheckoutWithCardsForm

    def charge_with_card(self, card: Card) -> Charge:
        """
        Charge the user with selected card.

        :returns: An instance of Charge object
        """
        customer = self.get_omise_customer()
        charge_details = self.get_charge_details()
        charge = customer.charge_with_card(
            amount=charge_details["amount"],
            currency=charge_details["currency"],
            card=card,
            **self.get_charge_kwargs(),
        )

        return charge

    def get_charge_details(self) -> Dict[int, Currency]:
        """
        Overwrite this method to provide the charge amount and currency.

        Read more on the amount and currency here https://www.omise.co/currency-and-amount

        :returns: Dictionary of charge amount in the smallest unit (int) and the 3-digit currency code. {"amount": int, "currency": str}
        """

    def get_omise_customer(self) -> Customer:
        """
        Overwrite this method to provide the customer instance to charge.

        :returns: An instance of Customer
        """
        if self.request.user.is_authenticated:
            customer, created = Customer.get_or_create(user=self.request.user)
            return customer

    def get_charge_kwargs(self) -> Dict:
        """
        Overwrite this method to add charge parameters.

        :return: A dictionary of charge parameters
        """
        return dict()

    def form_valid(self, form):

        selected_card = form.cleaned_data["card"]

        try:
            charge = self.charge_with_card(card=selected_card)
        except omise.errors.InvalidChargeError as e:
            messages.error(
                self.request,
                f"Payment failed: {str(e)}",
            )
            return super().form_invalid(form)

        self.process_new_charge(charge)

        if charge.status == ChargeStatus.FAILED:
            messages.error(
                self.request,
                f"Payment failed: {charge.failure_message}",
            )
            return super().form_invalid(form)

        if charge.status == ChargeStatus.PENDING:
            return redirect(charge.authorize_uri)

        if charge.status == ChargeStatus.SUCCESSFUL:
            messages.success(self.request, _("Payment successful."))

        return super().form_valid(form)

    def process_new_charge(self, charge: Charge):
        """
        Overwrite this method to process the new charge object.

        :params charge: The new Charge object.
        """
        pass



import json

from django.contrib.auth.mixins import LoginRequiredMixin
from django.contrib import messages
from django.contrib.messages.views import SuccessMessageMixin
from django.urls import reverse_lazy, reverse
from django.utils.translation import gettext_lazy as _
from django.shortcuts import redirect

from django.views.decorators.csrf import csrf_exempt
from django.views.generic import FormView, UpdateView, View, DetailView, TemplateView

from django.http import JsonResponse

from .forms import AddCardForm
from .mixins import CheckoutMixin
from .models.core import Customer, Card, Charge
from .models.event import Event, EventType
from .models.choices import ChargeStatus, Currency
from .omise import omise
from .utils.core_utils import update_or_create_from_omise_object
from .utils.event_utils import pre_event_handle, post_event_handle

from typing import Dict

# Create your views here.
@csrf_exempt
def omise_webhook_view(request):

    try:
        raw_event_data = json.loads(request.body)
    except json.decoder.JSONDecodeError:
        return JsonResponse(
            {"success": False, "message": "The data could not be parsed with JSON"},
            status=500,
        )

    try:
        omise_event = omise.Event.retrieve(raw_event_data.get("id"))
    except omise.errors.NotFoundError:
        return JsonResponse(
            {
                "success": False,
                "message": f"Could not find the event with the id {raw_event_data.get('id')}.",
            },
            status=404,
        )

    pre_event_handle(omise_event=omise_event, raw_event=raw_event_data)

    event, created = Event.objects.update_or_create(
        id=omise_event.id,
        livemode=omise_event.livemode,
        defaults={
            "event_type": omise_event.key,
            "date_created": omise_event.created_at,
            "data": raw_event_data,
        },
    )

    event_data = omise_event.data
    if omise_event.key not in [EventType.CARD_DESTROY.value]:
        event_data.reload()

    related_object = update_or_create_from_omise_object(omise_object=event_data)

    if related_object is not None:
        event.event_object = related_object
        event.save()

    post_event_handle(
        omise_event=omise_event, event_object=event, raw_event=raw_event_data
    )

    response = {}

    return JsonResponse(response, status=200)


class OmiseReturnURIView(View):

    redirect_url = "/"

    def get_redirect_url(self, charge):

        if charge.status == ChargeStatus.SUCCESSFUL:
            return self.request.GET.get(
                "successful_url",
                charge.metadata.get("successful_url", self.redirect_url),
            )

        elif charge.status == ChargeStatus.FAILED:
            return self.request.GET.get(
                "failed_url", charge.metadata.get("failed_url", self.redirect_url)
            )

        return self.redirect_url

    def get(self, request, uid):

        charge = Charge.objects.get(uid=uid)
        omise_charge = omise.Charge.retrieve(charge.id)
        charge = Charge.update_or_create_from_omise_object(omise_object=omise_charge)

        max_try = 5
        try_count = 0

        while charge.status == ChargeStatus.PENDING and try_count < max_try:
            omise_charge = omise.Charge.retrieve(charge.id)
            charge = Charge.update_or_create_from_omise_object(
                omise_object=omise_charge
            )
            try_count += 1

        if charge.status == ChargeStatus.SUCCESSFUL:
            messages.success(request, _("Payment successful"))

        if charge.status == ChargeStatus.FAILED:
            messages.warning(request, charge.failure_message)

        return redirect(self.get_redirect_url(charge))


class PromptpayCheckoutView(DetailView):
    model = Charge
    template_name = "django_omise/promptpay_checkout.html"
    context_object_name = "charge"

    def dispatch(self, *args, **kwargs):
        charge = self.get_object()
        source = charge.source

        if source.charge_status == ChargeStatus.PENDING:
            source = source.reload_from_omise()

        if source.charge_status != ChargeStatus.PENDING:
            return redirect(
                reverse("django_omise:return_uri", kwargs={"uid": charge.uid})
            )

        return super().dispatch(*args, **kwargs)


def charge_status_json(request):

    if request.method == "POST":

        try:
            request_body = json.loads(request.body)
        except json.decoder.JSONDecodeError:
            return JsonResponse(
                {"success": False, "message": "The data could not be parsed with JSON"},
                status=400,
            )

        charge_id = request_body.get("charge", None)

        if charge_id is None:
            return JsonResponse(
                {"success": False, "message": "Value 'charge' required"},
                status=400,
            )

        if not Charge.objects.filter(id=charge_id).exists():
            return JsonResponse(
                {"success": False, "message": "Object not found"}, status=404
            )

        charge = Charge.objects.get(id=charge_id)

        if charge.status == ChargeStatus.PENDING:
            charge = charge.reload_from_omise()

        return JsonResponse(
            {"success": True, "data": {"status": charge.status}},
        )

    else:
        return JsonResponse(
            {"success": False, "message": "Method not allowed"}, status=400
        )


class ManagePaymentMethodsView(LoginRequiredMixin, SuccessMessageMixin, FormView):

    form_class = AddCardForm
    template_name = "django_omise/manage_payment_methods.html"
    success_message = _("A new payment method has been added.")

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        context["customer"] = self.get_customer()
        return context

    def get_customer(self) -> Customer:
        """
        Get the customer to add the card to.

        Overwrite this method to change how the customer is retrieved.

        :returns: A Customer instance
        """
        customer, created = Customer.get_or_create(
            user=self.request.user,
        )

        return customer

    def form_valid(self, form):
        form.add_card_to_user(self.get_customer())
        return super().form_valid(form)

    def get_success_url(self, *args, **kwargs):
        return self.request.GET.get(
            "next", reverse_lazy("django_omise:manage_payment_methods")
        )


class PaymentMethodDeleteView(LoginRequiredMixin, UpdateView):

    model = Card
    fields = [
        "deleted",
    ]
    success_message = _("A card has been deleted.")
    success_url = reverse_lazy("django_omise:manage_payment_methods")
    template_name = "django_omise/card_confirm_delete.html"

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        context["customer"] = self.get_customer()
        return context

    def get_customer(self) -> Customer:
        """
        Get the customer to add the card to.

        Overwrite this method to change how the customer is retrieved.

        :returns: A Customer instance
        """
        customer, created = Customer.get_or_create(
            user=self.request.user,
        )
        return customer

    def form_valid(self, form):
        instance = form.save(commit=False)
        instance.delete()

        messages.success(self.request, self.success_message)
        return super().form_valid(form)

    def post(self, *args, **kwargs):

        object = self.get_object()
        if object.customer != self.get_customer():
            return redirect(reverse("django_omise:manage_payment_methods"))

        return super().post(*args, **kwargs)

    def get(self, *args, **kwargs):

        object = self.get_object()
        if object.customer != self.get_customer():
            return redirect(reverse("django_omise:manage_payment_methods"))

        return super().get(*args, **kwargs)


class CheckoutView(CheckoutMixin):
    template_name = "django_omise/checkout.html"

    def get_success_url(self, *args, **kwargs):
        return self.request.get_full_path()

    def get_charge_details(self) -> Dict[int, Currency]:
        """
        Overwrite this method to provide the charge amount and currency.

        Read more on the amount and currency here https://www.omise.co/currency-and-amount

        :returns: Dictionary of charge amount in the smallest unit (int) and the 3-digit currency code. {"amount": int, "currency": str}
        """
        return dict(
            amount=self.request.GET.get("amount", 100000),
            currency=self.request.GET.get("currency", "THB"),
        )

    def get_charge_kwargs(self) -> Dict:
        """
        Overwrite this method to add charge parameters.

        :return: A dictionary of charge parameters
        """
        return dict(
            metadata=dict(
                successful_url=self.request.get_full_path(),
                failed_url=self.request.get_full_path(),
            )
        )


class OmiseAccountAndCapabilityJsonView(TemplateView):

    template_name = "django_omise/account_and_compatibility.html"

    def dispatch(self, request, *args, **kwargs):

        if not self.request.user.is_authenticated or not self.request.user.is_superuser:
            return JsonResponse(
                {"message": "You have no permission to view this page"},
                status=403,
            )

        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)

        context["account"] = dict(omise.Account.retrieve().__dict__.get("_attributes"))

        context["capability"] = dict(
            omise.Capability.retrieve().__dict__.get("_attributes")
        )
        return context
